@@ -89,7 +89,7 @@ struct _CamelEwsFolderPrivate {
 	gboolean fetch_pending;
 	GMutex state_lock;
 	GCond fetch_cond;
-	GHashTable *uid_eflags;
+	GHashTable *fetching_uids;
 };
 
 static gboolean ews_delete_messages (CamelFolder *folder, const GSList *deleted_items, gboolean expunge, GCancellable *cancellable, GError **error);
@@ -488,6 +488,15 @@ ews_update_mgtrequest_mime_calendar_itemid (const gchar *mime_fname,
 	return mime_fname_new;
 }
 
+static void
+ews_fetch_cancellable_cancelled_cb (GCancellable *cancellable,
+				    GCond *fetch_cond)
+{
+	g_return_if_fail (fetch_cond != NULL);
+
+	g_cond_broadcast (fetch_cond);
+}
+
 static CamelMimeMessage *
 camel_ews_folder_get_message (CamelFolder *folder,
                               const gchar *uid,
@@ -497,7 +506,7 @@ camel_ews_folder_get_message (CamelFolder *folder,
 {
 	CamelEwsFolder *ews_folder;
 	CamelEwsFolderPrivate *priv;
-	EEwsConnection *cnc;
+	EEwsConnection *cnc = NULL;
 	EEwsAdditionalProps *add_props = NULL;
 	CamelEwsStore *ews_store;
 	const gchar *mime_content;
@@ -510,39 +519,60 @@ camel_ews_folder_get_message (CamelFolder *folder,
 	gchar *mime_fname_new = NULL;
 	GError *local_error = NULL;
 
+	g_return_val_if_fail (CAMEL_IS_EWS_FOLDER (folder), NULL);
+
 	ews_store = (CamelEwsStore *) camel_folder_get_parent_store (folder);
 	ews_folder = (CamelEwsFolder *) folder;
 	priv = ews_folder->priv;
 
-	if (!camel_ews_store_connected (ews_store, cancellable, error))
-		return NULL;
-
 	g_mutex_lock (&priv->state_lock);
 
-	/* If another thread is already fetching this message, wait for it */
-
-	/* FIXME: We might end up refetching a message anyway, if another
-	 * thread has already finished fetching it by the time we get to
-	 * this point in the code — ews_folder_get_message_sync() doesn't
-	 * hold any locks when it calls get_message_from_cache() and then
-	 * falls back to this function. */
-	if (g_hash_table_lookup (priv->uid_eflags, uid)) {
-		do {
-			g_cond_wait (&priv->fetch_cond, &priv->state_lock);
-		} while (g_hash_table_lookup (priv->uid_eflags, uid));
-
+	message = camel_ews_folder_get_message_from_cache (ews_folder, uid, cancellable, NULL);
+	if (message) {
 		g_mutex_unlock (&priv->state_lock);
 
-		message = camel_ews_folder_get_message_from_cache (ews_folder, uid, cancellable, error);
 		return message;
 	}
 
+	/* If another thread is already fetching this message, wait for it */
+	if (g_hash_table_lookup (priv->fetching_uids, uid)) {
+		gulong cancelled_handler_id = 0;
+
+		if (G_IS_CANCELLABLE (cancellable)) {
+			cancelled_handler_id = g_cancellable_connect (cancellable,
+				G_CALLBACK (ews_fetch_cancellable_cancelled_cb),
+				&priv->fetch_cond, NULL);
+		}
+
+		do {
+			g_cond_wait (&priv->fetch_cond, &priv->state_lock);
+		} while (g_hash_table_lookup (priv->fetching_uids, uid) &&
+			 !g_cancellable_is_cancelled (cancellable));
+
+		if (cancelled_handler_id && G_IS_CANCELLABLE (cancellable))
+			g_cancellable_disconnect (cancellable, cancelled_handler_id);
+
+		if (g_cancellable_set_error_if_cancelled (cancellable, error)) {
+			g_mutex_unlock (&priv->state_lock);
+			return NULL;
+		}
+
+		message = camel_ews_folder_get_message_from_cache (ews_folder, uid, cancellable, NULL);
+		if (message || g_cancellable_set_error_if_cancelled (cancellable, error)) {
+			g_mutex_unlock (&priv->state_lock);
+			return message;
+		}
+	}
+
 	/* Because we're using this as a form of mutex, we *know* that
 	 * we won't be inserting where an entry already exists. So it's
 	 * OK to insert uid itself, not g_strdup (uid) */
-	g_hash_table_insert (priv->uid_eflags, (gchar *) uid, (gchar *) uid);
+	g_hash_table_insert (priv->fetching_uids, (gchar *) uid, (gchar *) uid);
 	g_mutex_unlock (&priv->state_lock);
 
+	if (!camel_ews_store_connected (ews_store, cancellable, error))
+		goto exit;
+
 	cnc = camel_ews_store_ref_connection (ews_store);
 	ids = g_slist_append (ids, (gchar *) uid);
 
@@ -554,13 +584,14 @@ camel_ews_folder_get_message (CamelFolder *folder,
 	    g_mkdir_with_parents (mime_dir, 0700) == -1) {
 		g_set_error (
 			error, CAMEL_ERROR, CAMEL_ERROR_GENERIC,
-			_("Unable to create cache path"));
+			_("Unable to create cache path '%s': %s"),
+			mime_dir, g_strerror (errno));
 		g_free (mime_dir);
 		goto exit;
 	}
 
 	add_props = e_ews_additional_props_new ();
-	add_props->field_uri = g_strdup ("item:MimeContent");
+	add_props->field_uri = g_strdup ("item:MimeContent message:From message:Sender");
 
 	res = e_ews_connection_get_items_sync (
 		cnc, pri, ids, "IdOnly", add_props,
@@ -645,7 +676,8 @@ camel_ews_folder_get_message (CamelFolder *folder,
 	if (g_mkdir_with_parents (dir, 0700) == -1) {
 		g_set_error (
 			error, CAMEL_ERROR, CAMEL_ERROR_GENERIC,
-			_("Unable to create cache path"));
+			_("Unable to create cache path '%s': %s"),
+			dir, g_strerror (errno));
 		g_free (dir);
 		g_free (cache_file);
 		goto exit;
@@ -655,21 +687,61 @@ camel_ews_folder_get_message (CamelFolder *folder,
 	if (g_rename (mime_content, cache_file) != 0) {
 		g_set_error (
 			error, CAMEL_ERROR, CAMEL_ERROR_GENERIC,
-			_("Failed to move message cache file"));
+			/* Translators: The first %s consists of the source file name,
+			   the second %s of the destination file name and
+			   the third %s of the error message. */
+			_("Failed to move message cache file from '%s' to '%s': %s"),
+			mime_content, cache_file, g_strerror (errno));
 		g_free (cache_file);
 		goto exit;
 	}
 	g_free (cache_file);
 
 	message = camel_ews_folder_get_message_from_cache (ews_folder, uid, cancellable, error);
+	if (message) {
+		CamelInternetAddress *from;
+		const gchar *email = NULL;
+
+		from = camel_mime_message_get_from (message);
+
+		if (!from || !camel_internet_address_get (from, 0, NULL, &email) || !email || !*email) {
+			const EwsMailbox *mailbox;
+
+			mailbox = e_ews_item_get_from (items->data);
+			if (!mailbox)
+				mailbox = e_ews_item_get_sender (items->data);
+			if (mailbox) {
+				CamelStream *cache_stream;
+
+				email = NULL;
+
+				if (g_strcmp0 (mailbox->routing_type, "EX") == 0)
+					email = e_ews_item_util_strip_ex_address (mailbox->email);
+
+				from = camel_internet_address_new ();
+				camel_internet_address_add (from, mailbox->name, email ? email : mailbox->email);
+				camel_mime_message_set_from (message, from);
+				g_object_unref (from);
+
+				g_rec_mutex_lock (&priv->cache_lock);
+				/* Ignore errors here, it's nothing fatal in this case */
+				cache_stream = ews_data_cache_get (ews_folder->cache, "cur", uid, NULL);
+				if (cache_stream) {
+					camel_data_wrapper_write_to_stream_sync (CAMEL_DATA_WRAPPER (message), cache_stream, cancellable, NULL);
+					g_object_unref (cache_stream);
+				}
+				g_rec_mutex_unlock (&priv->cache_lock);
+			}
+		}
+	}
 
 exit:
 	g_mutex_lock (&priv->state_lock);
-	g_hash_table_remove (priv->uid_eflags, uid);
-	g_mutex_unlock (&priv->state_lock);
+	g_hash_table_remove (priv->fetching_uids, uid);
 	g_cond_broadcast (&priv->fetch_cond);
+	g_mutex_unlock (&priv->state_lock);
 
-	if (!message && !error)
+	if (!message && error && !*error)
 		g_set_error (
 			error, CAMEL_ERROR, 1,
 			"Could not retrieve the message");
@@ -728,12 +800,11 @@ ews_folder_get_message_sync (CamelFolder *folder,
 {
 	CamelMimeMessage *message;
 
-	message = camel_ews_folder_get_message_from_cache ((CamelEwsFolder *) folder, uid, cancellable, NULL);
-	if (!message) {
-		message = camel_ews_folder_get_message (folder, uid, EWS_ITEM_HIGH, cancellable, error);
-		if (message)
-			ews_folder_maybe_update_mlist (folder, uid, message);
-	}
+	g_return_val_if_fail (CAMEL_IS_EWS_FOLDER (folder), NULL);
+
+	message = camel_ews_folder_get_message (folder, uid, EWS_ITEM_HIGH, cancellable, error);
+	if (message)
+		ews_folder_maybe_update_mlist (folder, uid, message);
 
 	return message;
 }
@@ -1124,7 +1195,7 @@ ews_move_to_junk_folder (CamelFolder *folder,
 		folder_id = camel_ews_store_summary_get_folder_id_from_folder_type (
 			ews_store->summary, CAMEL_FOLDER_TYPE_JUNK);
 
-		status = e_ews_connection_move_items_sync (
+		status = e_ews_connection_move_items_in_chunks_sync (
 			cnc, EWS_PRIORITY_MEDIUM, folder_id, FALSE,
 			junk_uids, &moved_items, cancellable, &local_error);
 
@@ -1136,17 +1207,20 @@ ews_move_to_junk_folder (CamelFolder *folder,
 			status = ews_refresh_info_sync (folder, cancellable, &local_error);
 		}
 
-		g_slist_free_full (moved_items, g_object_unref);
-		g_free (folder_id);
-
-		if (status) {
+		/* Messages could be moved partially only, like when
+		   the user cancels the operation in the middle */
+		if (status || moved_items) {
 			CamelFolderChangeInfo *changes;
-			const GSList *iter;
+			const GSList *iter, *items_iter;
 
 			changes = camel_folder_change_info_new ();
 
-			for (iter = junk_uids; iter; iter = g_slist_next (iter)) {
+			for (iter = junk_uids, items_iter = moved_items; iter && items_iter; iter = g_slist_next (iter), items_iter = g_slist_next (items_iter)) {
 				const gchar *uid = iter->data;
+				EEwsItem *item = items_iter->data;
+
+				if (!item || e_ews_item_get_item_type (item) == E_EWS_ITEM_TYPE_ERROR)
+					continue;
 
 				camel_folder_summary_lock (folder->summary);
 
@@ -1164,6 +1238,9 @@ ews_move_to_junk_folder (CamelFolder *folder,
 			camel_folder_change_info_free (changes);
 		}
 
+		g_slist_free_full (moved_items, g_object_unref);
+		g_free (folder_id);
+
 		if (local_error) {
 			camel_ews_store_maybe_disconnect (ews_store, local_error);
 			g_propagate_error (error, local_error);
@@ -2005,13 +2082,17 @@ ews_transfer_messages_to_sync (CamelFolder *source,
 
 	ids = g_slist_reverse (ids);
 
-	if (success && e_ews_connection_move_items_sync (
+	success = success && e_ews_connection_move_items_in_chunks_sync (
 		cnc, EWS_PRIORITY_MEDIUM,
 		dst_id, !delete_originals,
 		ids, &ret_items,
-		cancellable, &local_error)) {
+		cancellable, &local_error);
+
+	/* Messages could be copied/moved partially only, like when
+	   the user cancels the operation in the middle */
+	if (success || ret_items) {
 		CamelFolderChangeInfo *changes;
-		GSList *l;
+		GSList *l, *processed_items = NULL;
 
 		changes = camel_folder_change_info_new ();
 
@@ -2022,10 +2103,14 @@ ews_transfer_messages_to_sync (CamelFolder *source,
 			CamelMessageInfo *clone;
 			const EwsId *id;
 
-			if (e_ews_item_get_item_type (l->data) == E_EWS_ITEM_TYPE_ERROR)
+			if (e_ews_item_get_item_type (l->data) == E_EWS_ITEM_TYPE_ERROR) {
+				if (!local_error)
+					local_error = g_error_copy (e_ews_item_get_error (l->data));
 				continue;
+			}
 
 			id = e_ews_item_get_id (l->data);
+			processed_items = g_slist_prepend (processed_items, uids->pdata[i]);
 
 			message = ews_folder_get_message_cached (source, uids->pdata[i], cancellable);
 			if (message == NULL)
@@ -2069,10 +2154,12 @@ ews_transfer_messages_to_sync (CamelFolder *source,
 		if (delete_originals) {
 			changes = camel_folder_change_info_new ();
 
-			for (i = 0; i < uids->len; i++) {
-				camel_folder_summary_remove_uid (source->summary, uids->pdata[i]);
-				camel_folder_change_info_remove_uid (changes, uids->pdata[i]);
-				ews_data_cache_remove (CAMEL_EWS_FOLDER (source)->cache, "cur", uids->pdata[i], NULL);
+			for (l = processed_items; l; l = g_slist_next (l)) {
+				const gchar *uid = l->data;
+
+				camel_folder_summary_remove_uid (source->summary, uid);
+				camel_folder_change_info_remove_uid (changes, uid);
+				ews_data_cache_remove (CAMEL_EWS_FOLDER (source)->cache, "cur", uid, NULL);
 			}
 			if (camel_folder_change_info_changed (changes)) {
 				camel_folder_summary_touch (source->summary);
@@ -2085,8 +2172,13 @@ ews_transfer_messages_to_sync (CamelFolder *source,
 		/* update destination folder only if not frozen, to not update
 		   for each single message transfer during filtering
 		 */
-		if (!camel_folder_is_frozen (destination))
+		if (!camel_folder_is_frozen (destination)) {
+			camel_operation_progress (cancellable, -1);
+
 			ews_refresh_info_sync (destination, cancellable, NULL);
+		}
+
+		g_slist_free (processed_items);
 	}
 	g_free (dst_id);
 
@@ -2415,7 +2507,7 @@ ews_folder_finalize (GObject *object)
 	g_mutex_clear (&ews_folder->priv->search_lock);
 	g_mutex_clear (&ews_folder->priv->state_lock);
 	g_rec_mutex_clear (&ews_folder->priv->cache_lock);
-	g_hash_table_destroy (ews_folder->priv->uid_eflags);
+	g_hash_table_destroy (ews_folder->priv->fetching_uids);
 	g_cond_clear (&ews_folder->priv->fetch_cond);
 
 	/* Chain up to parent's finalize() method. */
@@ -2510,7 +2602,7 @@ camel_ews_folder_init (CamelEwsFolder *ews_folder)
 	ews_folder->priv->refreshing = FALSE;
 
 	g_cond_init (&ews_folder->priv->fetch_cond);
-	ews_folder->priv->uid_eflags = g_hash_table_new (g_str_hash, g_str_equal);
+	ews_folder->priv->fetching_uids = g_hash_table_new (g_str_hash, g_str_equal);
 	camel_folder_set_lock_async (folder, TRUE);
 }
 
